package net.wanji.web.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.util.ObjectUtil;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import net.wanji.common.gts.GeometryUtil;
import net.wanji.common.gts.GeomsConvertUtil;
import net.wanji.common.gts.service.GtsService;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.dao.entity.RidInfoEntity;
import net.wanji.databus.dao.mapper.BaseCrossInfoMapper;
import net.wanji.databus.dao.mapper.CrossDirDataRealtimeMapper;
import net.wanji.databus.dao.mapper.RidInfoMapper;
import net.wanji.databus.po.BaseCrossInfoPO;
import net.wanji.web.bo.AddSpecialServiceBO;
import net.wanji.web.bo.IsValidPointBO;
import net.wanji.web.bo.SpecialServiceRouteBO;
import net.wanji.web.common.enums.CrossDirEnum;
import net.wanji.web.mapper.*;
import net.wanji.web.po.CrossDirTurnPO;
import net.wanji.web.po.RidInfoPO;
import net.wanji.web.po.SpecialServiceCrossPO;
import net.wanji.web.po.SpecialServicePO;
import net.wanji.web.service.SpecialServiceService;
import net.wanji.web.vo.RouteElementVO;
import net.wanji.web.vo.specialService.DeleteCrossInVO;
import net.wanji.web.vo.specialService.ListSpecialServicesOutVO;
import net.wanji.web.vo.specialService.UpdateCrossInVO;
import net.wanji.web.vo.specialService.UpdateSpecialServiceInVO;
import org.jetbrains.annotations.NotNull;
import org.locationtech.jts.geom.Geometry;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.*;

/**
 * @author Kent HAN
 * @date 2022/11/8 11:07
 */
@Service
public class SpecialServiceServiceImpl implements SpecialServiceService {
    @Autowired
    SpecialServiceMapper specialServiceMapper;
    @Autowired
    SpecialServiceCrossMapper specialServiceCrossMapper;
    @Autowired
    CustomRidInfoMapper customRidInfoMapper;
    @Autowired
    BaseCrossInfoMapper tBaseBaseCrossInfoMapper;
    @Autowired
    CrossDirTurnMapper crossDirTurnMapper;
    @Autowired
    CrossPhasePlanMapper crossPhasePlanMapper;
    @Autowired
    private BaseCrossInfoMapper baseCrossInfoMapper;
    @Autowired
    private RidInfoMapper ridInfoMapper;

    private Gson gson = new GsonBuilder().setDateFormat("yyyy-MM-dd HH:mm:ss").create();
    @Autowired
    private GtsService gtsService;
    @Autowired
    private CrossDirDataRealtimeMapper crossDirDataRealtimeMapper;

    @Override
    @Transactional
    public void addSpecialService(AddSpecialServiceBO addSpecialServiceBO) {
        // 插入特勤表
        SpecialServicePO specialServicePO = new SpecialServicePO();
        specialServicePO.setName(addSpecialServiceBO.getName());
        specialServicePO.setPlateNum(addSpecialServiceBO.getPlateNum());
        specialServicePO.setStartLocation(addSpecialServiceBO.getStartLocation());
        specialServicePO.setEndLocation(addSpecialServiceBO.getEndLocation());

        specialServiceMapper.insertOne(specialServicePO);
        Integer specialServiceId = specialServicePO.getId();

        // 插入特勤路口关系表
        List<RouteElementVO> route = addSpecialServiceBO.getRoute();

        for (int i = 0; i < route.size(); i++) {
            RouteElementVO routeElementVO = route.get(i);
            String startName = routeElementVO.getStartName();
            BaseCrossInfoPO startCross = baseCrossInfoMapper.selectByName(startName);
            String startCrossId = startCross.getId();
            String endName = routeElementVO.getEndName();
            BaseCrossInfoPO endCross = baseCrossInfoMapper.selectByName(endName);
            String endCrossId = endCross.getId();


            RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(startCrossId, endCrossId);
            if (i != route.size() - 1) { // 非结束路段
                SpecialServiceCrossPO specialServiceCrossPO = new SpecialServiceCrossPO();
                specialServiceCrossPO.setSpecialServiceId(specialServiceId);
                specialServiceCrossPO.setCrossId(startCrossId);
                // 计算方向
                if (i == 0) { // 开始路口
                    specialServiceCrossPO.setInDir(0);
                    Integer outDir = ridInfoEntity.getOutDir();
                    specialServiceCrossPO.setOutDir(outDir);
                } else { // 中间路口
                    // 获取上一个路段
                    RouteElementVO preRoute = route.get(i - 1);
                    String preStartName = preRoute.getStartName();
                    BaseCrossInfoPO preStartCross = baseCrossInfoMapper.selectByName(preStartName);
                    String preStartCrossId = preStartCross.getId();

                    String preEndName = preRoute.getEndName();
                    BaseCrossInfoPO preEndCross = baseCrossInfoMapper.selectByName(preEndName);
                    String preEndCrossId = preEndCross.getId();
                    RidInfoEntity preRid = ridInfoMapper.selectByStartEnd(preStartCrossId, preEndCrossId);
                    // 当前路口驶入方向
                    Integer inDir = preRid.getInDir();
                    // 当前路口使出方向
                    Integer outDir = ridInfoEntity.getOutDir();
                    specialServiceCrossPO.setInDir(inDir);
                    specialServiceCrossPO.setOutDir(outDir);
                }
                specialServiceCrossPO.setSort(i + 1);
                specialServiceCrossPO.setStatus(0);
                specialServiceCrossPO.setAutoUnlock(0);
                specialServiceCrossMapper.insertOne(specialServiceCrossPO);
            } else { // 结束路段
                // 开始路口
                SpecialServiceCrossPO specialServiceCrossPO = new SpecialServiceCrossPO();
                specialServiceCrossPO.setSpecialServiceId(specialServiceId);
                specialServiceCrossPO.setCrossId(startCrossId);
                // 获取上一个路段
                RouteElementVO preRoute = route.get(i - 1);
                String preStartName = preRoute.getStartName();
                BaseCrossInfoPO preStartCross = baseCrossInfoMapper.selectByName(preStartName);
                String preStartCrossId = preStartCross.getId();

                String preEndName = preRoute.getEndName();
                BaseCrossInfoPO preEndCross = baseCrossInfoMapper.selectByName(preEndName);
                String preEndCrossId = preEndCross.getId();
                RidInfoEntity preRid = ridInfoMapper.selectByStartEnd(preStartCrossId, preEndCrossId);
                // 当前路口驶入方向
                Integer inDir = preRid.getInDir();
                // 当前路口使出方向
                Integer outDir = ridInfoEntity.getOutDir();
                specialServiceCrossPO.setInDir(inDir);
                specialServiceCrossPO.setOutDir(outDir);
                specialServiceCrossPO.setSort(i + 1);
                specialServiceCrossPO.setStatus(0);
                specialServiceCrossPO.setAutoUnlock(0);
                specialServiceCrossMapper.insertOne(specialServiceCrossPO);
                // 结束路口
                SpecialServiceCrossPO specialServiceCrossPOEnd = new SpecialServiceCrossPO();
                specialServiceCrossPOEnd.setSpecialServiceId(specialServiceId);
                specialServiceCrossPOEnd.setCrossId(endCrossId);
                Integer inDirEnd = ridInfoEntity.getInDir();
                specialServiceCrossPOEnd.setInDir(inDirEnd);
                specialServiceCrossPOEnd.setOutDir(0);
                specialServiceCrossPOEnd.setSort(i + 2);
                specialServiceCrossPOEnd.setStatus(0);
                specialServiceCrossPOEnd.setAutoUnlock(0);
                specialServiceCrossMapper.insertOne(specialServiceCrossPOEnd);
            }
        }
    }

    @Override
    public List<ListSpecialServicesOutVO> listSpecialServices(String adCode) {
        List<SpecialServicePO> specialServicePOList = specialServiceMapper.selectByAdCode(adCode);
        List<ListSpecialServicesOutVO> listSpecialServicesOutVOList = new ArrayList<>();
        for (SpecialServicePO specialServicePO : specialServicePOList) {
            ListSpecialServicesOutVO listSpecialServicesOutVO = new ListSpecialServicesOutVO();
            BeanUtils.copyProperties(specialServicePO, listSpecialServicesOutVO);
            // 构造wkt列表
            List<String> wkts = new ArrayList<>();
            Integer specialServiceId = specialServicePO.getId();
            List<SpecialServiceCrossPO> specialServiceCrossPOList =
                    specialServiceCrossMapper.selectBySpecialServiceId(specialServiceId);
            int size = specialServiceCrossPOList.size();
            for (int i = 0; i < size - 1; i++) {
                SpecialServiceCrossPO currentCross = specialServiceCrossPOList.get(i);
                SpecialServiceCrossPO nextCross = specialServiceCrossPOList.get(i + 1);
                RidInfoPO ridInfoPO = customRidInfoMapper.selectOne(currentCross.getCrossId(), nextCross.getCrossId());
                String wkt = ridInfoPO.getWkt();
                wkts.add(wkt);
            }
            listSpecialServicesOutVO.setWkts(wkts);
            listSpecialServicesOutVOList.add(listSpecialServicesOutVO);
        }
        return listSpecialServicesOutVOList;
    }

    @Override
    public void updateSpecialService(UpdateSpecialServiceInVO updateSpecialServiceInVO) {
        specialServiceMapper.updateSpecialService(updateSpecialServiceInVO);
    }

    @Override
    @Transactional
    public void deleteSpecialService(Integer id) {
        specialServiceMapper.deleteSpecialService(id);
        specialServiceCrossMapper.deleteCrossBySpecialServiceId(id);
    }

    @Override
    @Transactional
    public void updateCross(UpdateCrossInVO updateCrossInVO) {
        Integer specialServiceId = updateCrossInVO.getSpecialServiceId();
        String crossId = updateCrossInVO.getCrossId();
        String dirInAndOut = updateCrossInVO.getDirInAndOut();
        // 获取方向（东西南北...）
        int length = dirInAndOut.length();
        String dirStr = dirInAndOut.substring(0, length - 2);
        Integer dir = CrossDirEnum.getCodeByMsg(dirStr);
        specialServiceCrossMapper.updatDir(specialServiceId, crossId, dir);
        // 更新当前路口转向
        Integer sort = updateCrossInVO.getSort();
        Integer nextSort = sort + 1;
        String endCrossid = specialServiceCrossMapper.selectCrossIdBySort(specialServiceId, nextSort);
        String startCrossOutDirId = customRidInfoMapper.selectStartCrossOutDir(crossId, endCrossid);
        String startCrossInDirId = crossId + "_" + dir + "_2_1";
        CrossDirTurnPO crossDirTurnPO = crossDirTurnMapper.selectOneByInAndOutIds(startCrossInDirId, startCrossOutDirId);
        String turntype = crossDirTurnPO.getTurntype();
        specialServiceCrossMapper.updateTurn(specialServiceId, crossId, turntype);
    }

    @Override
    @Transactional
    public void deleteCross(DeleteCrossInVO deleteCrossInVO) {
        String crossId = deleteCrossInVO.getCrossId();
        Integer specialServiceId = deleteCrossInVO.getSpecialServiceId();
        Integer sort = deleteCrossInVO.getSort();
        if (sort > 1) {
            Integer preSort = sort - 1;
            String preCrossid = specialServiceCrossMapper.selectCrossIdBySort(specialServiceId, preSort);
            // 减少路线长度
            Double roadLength = customRidInfoMapper.selectRoadLength(preCrossid, crossId);
            Double oldLength = specialServiceMapper.selectRoadLength(specialServiceId);
            oldLength -= roadLength;
            specialServiceMapper.updateLength(specialServiceId, oldLength);
            // 删除上一个路口转向
            specialServiceCrossMapper.updateTurn(specialServiceId, preCrossid, null);
        }
        specialServiceCrossMapper.deleteCross(specialServiceId, crossId);
    }

    @Override
    public Set<String> crossInList(String crossId) {
        List<String> crossInIdList = crossDirTurnMapper.crossInList(crossId);
        Set<String> res = new HashSet<>();
        for (String crossInId : crossInIdList) {
            String crossIn = getCrossIn(crossInId);
            res.add(crossIn);
        }
        return res;
    }

    private String getCrossIn(String crossInId) {
        // 从 13FNK0C6790_1_2_1 中获取 “北进口”
        int length = crossInId.length();
        String dir = crossInId.substring(length - 5, length - 4);
        String dirStr = CrossDirEnum.getMsgByCode(Integer.parseInt(dir));
        String s = dirStr + "进口";
        return s;
    }

    @Override
    public List<List<RouteElementVO>> specialServiceRoute(SpecialServiceRouteBO specialServiceRouteBO) {
        List<RidInfoEntity> ridInfoEntities = ridInfoMapper.selectAll();
        // 获取起点经纬度与rid连线相交的路段
        String startLonLat = specialServiceRouteBO.getStartLonLat();
        Geometry startGeometry = GeometryUtil.str2Geometry(startLonLat);
        Set<String> startRidPoints = new HashSet<>();
        for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
            String ridLine = ridInfoEntity.getWkt();
            Geometry ridLineGeometry = GeometryUtil.str2Geometry(ridLine);
            Geometry ridLineBuf = GeomsConvertUtil.getGeometryBuf(ridLineGeometry, "50.0");
            boolean isIntersect = ridLineBuf.intersects(startGeometry);
            if (isIntersect) {
                String startCrossId = ridInfoEntity.getStartCrossId();
                String endCrossId = ridInfoEntity.getEndCrossId();
                startRidPoints.add(startCrossId);
                startRidPoints.add(endCrossId);
            }
        }
        // 获取起点路口
        double startMinDistance = Double.MAX_VALUE;
        BaseCrossInfoPO startCross = null;
        for (String ridPoint : startRidPoints) {
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(ridPoint);
            String location = baseCrossInfoPO.getLocation();
            double[] lonLat = CrossUtil.getLonLat(location);
            String ridPointLonLat = lonLat[0] + "," + lonLat[1];
            double distance = gtsService.distance(startLonLat, ridPointLonLat);
            startMinDistance = Math.min(startMinDistance, distance);
            if (distance == startMinDistance) {
                startCross = baseCrossInfoPO;
            }
        }

        // 获取终点经纬度与rid连线相交的路段
        String endLonLat = specialServiceRouteBO.getEndLonLat();
        Geometry endGeometry = GeometryUtil.str2Geometry(endLonLat);
        Set<String> endRidPoints = new HashSet<>();
        for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
            String ridLine = ridInfoEntity.getWkt();
            Geometry ridLineGeometry = GeometryUtil.str2Geometry(ridLine);
            Geometry ridLineBuf = GeomsConvertUtil.getGeometryBuf(ridLineGeometry, "50.0");
            boolean isIntersect = ridLineBuf.intersects(endGeometry);
            if (isIntersect) {
                String startCrossId = ridInfoEntity.getStartCrossId();
                String endCrossId = ridInfoEntity.getEndCrossId();
                endRidPoints.add(startCrossId);
                endRidPoints.add(endCrossId);
            }
        }
        // 获取终点路口
        double endMinDistance = Double.MAX_VALUE;
        BaseCrossInfoPO endCross = null;
        for (String ridPoint : endRidPoints) {
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(ridPoint);
            String location = baseCrossInfoPO.getLocation();
            double[] lonLat = CrossUtil.getLonLat(location);
            String ridPointLonLat = lonLat[0] + "," + lonLat[1];
            double distance = gtsService.distance(endLonLat, ridPointLonLat);
            endMinDistance = Math.min(endMinDistance, distance);
            if (distance == endMinDistance) {
                endCross = baseCrossInfoPO;
            }
        }

        if (ObjectUtil.isNotEmpty(startCross) && ObjectUtil.isNotEmpty(endCross)
                && ObjectUtil.notEqual(startCross, endCross)) {
            List<List<RouteElementVO>> res = new ArrayList<>();
            // 构造第一条备选线路
            List<BaseCrossInfoPO> routeList1 = getRouteList(startCross, endCross);
            List<RouteElementVO> route1 = buildRoute(routeList1);
            // 构造第二条备选线路
            String startCrossId = startCross.getId();
            String endCrossLonLatStr = getCrossLonLatStr(endCross);
            List<RidInfoEntity> ridUpCrossList = ridInfoMapper.selectUpCross(startCrossId);
            Map<Double, BaseCrossInfoPO> distanceMap = new HashMap<>();
            List<Double> distanceList = new ArrayList<>();
            for (RidInfoEntity ridUpCross : ridUpCrossList) {
                String upCrossId = ridUpCross.getStartCrossId();
                BaseCrossInfoPO upCross = baseCrossInfoMapper.selectById(upCrossId);
                String upCrossLonLatStr = getCrossLonLatStr(upCross);
                double distance = gtsService.distance(endCrossLonLatStr, upCrossLonLatStr);
                distanceMap.put(distance, upCross);
                distanceList.add(distance);
            }
            Collections.sort(distanceList);
            BaseCrossInfoPO secondStartCross = null;
            try {
                secondStartCross = distanceMap.get(distanceList.get(1));
            } catch (IndexOutOfBoundsException e) {
                throw new RuntimeException("路网数据不连续");
            }
            List<BaseCrossInfoPO> routeList2 = new ArrayList<>();
            routeList2.add(startCross);
            getSecondRouteList(secondStartCross, endCross, routeList2);
            List<RouteElementVO> route2 = buildRoute(routeList2);

            res.add(route1);
            res.add(route2);
            return res;
        }
        return null;
    }

    private void getSecondRouteList(BaseCrossInfoPO secondStartCross, BaseCrossInfoPO endCross,
                                    List<BaseCrossInfoPO> routeList2) {
        String endCrossLonLatStr = getCrossLonLatStr(endCross);
        // 获取上游路口到终点距离最近的
        BaseCrossInfoPO nextStartCross = secondStartCross;
        while (ObjectUtil.notEqual(endCross.getId(), nextStartCross.getId())) {
            String nextCrossId = nextStartCross.getId();
            double minDistance = Double.MAX_VALUE;
            List<RidInfoEntity> ridUpCrossList = ridInfoMapper.selectUpCross(nextStartCross.getId());
            for (RidInfoEntity ridUpCross : ridUpCrossList) {
                String upCrossId = ridUpCross.getStartCrossId();
                BaseCrossInfoPO upCross = baseCrossInfoMapper.selectById(upCrossId);
                String upCrossLonLatStr = getCrossLonLatStr(upCross);
                double distance = gtsService.distance(endCrossLonLatStr, upCrossLonLatStr);
                minDistance = Math.min(minDistance, distance);
                if (minDistance == distance && !hasCross(routeList2, upCross)) { // 已选过的路口不可再选
                    nextStartCross = upCross;
                }
            }
            if (ObjectUtil.equal(nextStartCross.getId(), nextCrossId)) {
                throw new RuntimeException("路网数据不连续");
            }
            routeList2.add(nextStartCross);
        }
        routeList2.add(1, secondStartCross);
    }

    private boolean hasCross(List<BaseCrossInfoPO> routeList, BaseCrossInfoPO upCross) {
        for (BaseCrossInfoPO baseCrossInfoPO : routeList) {
            String id = baseCrossInfoPO.getId();
            String upCrossId = upCross.getId();
            if (ObjectUtil.equal(id, upCrossId)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void isValidPoint(IsValidPointBO isValidPointBO) {
        String lonLat = isValidPointBO.getLonLat();
        List<RidInfoEntity> ridInfoEntities = ridInfoMapper.selectAll();
        // 获取点位经纬度与rid连线相交的路段
        Geometry geometry = GeometryUtil.str2Geometry(lonLat);
        Set<String> startRidPoints = new HashSet<>();
        for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
            String ridLine = ridInfoEntity.getWkt();
            Geometry ridLineGeometry = GeometryUtil.str2Geometry(ridLine);
            Geometry ridLineBuf = GeomsConvertUtil.getGeometryBuf(ridLineGeometry, "50.0");
            boolean isIntersect = ridLineBuf.intersects(geometry);
            if (isIntersect) {
                String startCrossId = ridInfoEntity.getStartCrossId();
                String endCrossId = ridInfoEntity.getEndCrossId();
                startRidPoints.add(startCrossId);
                startRidPoints.add(endCrossId);
            }
        }
        if (CollectionUtil.isEmpty(startRidPoints)) {
            throw new RuntimeException("不是有效点位");
        }
    }

    @NotNull
    private List<BaseCrossInfoPO> getRouteList(BaseCrossInfoPO startCross, BaseCrossInfoPO endCross) {
        String endCrossLonLatStr = getCrossLonLatStr(endCross);
        // 获取上游路口到终点距离最近的
        BaseCrossInfoPO nextStartCross = startCross;
        List<BaseCrossInfoPO> routeList = new ArrayList<>();
        while (ObjectUtil.notEqual(endCross.getId(), nextStartCross.getId())) {
            String nextCrossId = nextStartCross.getId();
            double minDistance = Double.MAX_VALUE;
            List<RidInfoEntity> ridUpCrossList = ridInfoMapper.selectUpCross(nextStartCross.getId());
            for (RidInfoEntity ridUpCross : ridUpCrossList) {
                String upCrossId = ridUpCross.getStartCrossId();
                BaseCrossInfoPO upCross = baseCrossInfoMapper.selectById(upCrossId);
                String upCrossLonLatStr = getCrossLonLatStr(upCross);
                double distance = gtsService.distance(endCrossLonLatStr, upCrossLonLatStr);
                minDistance = Math.min(minDistance, distance);
                if (minDistance == distance && !hasCross(routeList, upCross)) { // 已选过的路口不可再选
                    nextStartCross = upCross;
                }
            }
            if (ObjectUtil.equal(nextStartCross.getId(), nextCrossId)) {
                throw new RuntimeException("路网数据不连续");
            }
            routeList.add(nextStartCross);
        }
        routeList.add(0, startCross);
        return routeList;
    }

    private List<RouteElementVO> buildRoute(List<BaseCrossInfoPO> routeList) {
        List<RouteElementVO> route = new ArrayList<>();
        // 遍历到倒数第二个路口
        for (int i = 0; i < routeList.size() - 1; i++) {
            RouteElementVO routeElementVO = new RouteElementVO();
            BaseCrossInfoPO firstCross = routeList.get(i);
            BaseCrossInfoPO secondCross = routeList.get(i + 1);
            String firstLonLatStr = getCrossLonLatStr(firstCross);
            String secondLonLatStr = getCrossLonLatStr(secondCross);

            routeElementVO.setStartLonLat(firstLonLatStr);
            routeElementVO.setStartName(firstCross.getName());
            routeElementVO.setEndLonLat(secondLonLatStr);
            routeElementVO.setEndName(secondCross.getName());

            // 获取路况，1为进口
            Integer status = getRidStatus(firstCross, secondCross, 1);
            routeElementVO.setStatus(status);

            route.add(routeElementVO);
        }
        return route;
    }

    private Integer getRidStatus(BaseCrossInfoPO firstCross, BaseCrossInfoPO secondCross, int inOutType) {
        String firstId = firstCross.getId();
        String secondId = secondCross.getId();
        RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(firstId, secondId);
        Integer inDir = ridInfoEntity.getInDir();
        Integer status = crossDirDataRealtimeMapper.selectStatus(secondId, inDir, inOutType);
        return status;
    }

    private String getCrossLonLatStr(BaseCrossInfoPO firstCross) {
        String firstLocation = firstCross.getLocation();
        double[] firstLonLat = CrossUtil.getLonLat(firstLocation);
        return firstLonLat[0] + "," + firstLonLat[1];
    }
}